/*
     eib.c - Low level EIB driver for BCU 1

     Copyright (C) 2000 Bernd Thallner <bernd@kangaroo.at>
     Copyright (C) 2005-2006 Martin Kgler <mkoegler@auto.tuwien.ac.at>

     Ported to Linux 2.6 by Robert Kurevija
     Heavily modfied and restructed by Martin Kgler

     This program is free software; you can redistribute it and/or
     modify it under the terms of the GNU General Public License
     as published by the Free Software Foundation; either version 2
     of the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/device.h>
#include <linux/version.h>
#include <linux/serial_reg.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
#else
#include <linux/platform_device.h>
#endif

#include"eib.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
#define HAVE_SIMPLE_CLASS
#else
#undef HAVE_SIMPLE_CLASS
#endif

#define EIB_DEBUG

#undef EIB_PRINT
#define EIB_PRINT(fmt, args...) printk(KERN_NOTICE "eib: " fmt "\n", ## args)

#undef PDEBUG
#ifdef EIB_DEBUG
#  define PDEBUG(fmt, args...) printk(KERN_NOTICE "eib: " fmt "\n", ## args)
#else
# define PDEBUG(fmt, args...) do {} while (0)
#endif

#define MAX_DEVICE_COUNT 4
#define DEFAULT_MAJOR 121

static int eib_major = DEFAULT_MAJOR;

MODULE_AUTHOR ("Martin Kgler, mkoegler@auto.tuwien.ac.at");
MODULE_AUTHOR ("Bernd Thallner, bernd@kangaroo.at");
MODULE_DESCRIPTION ("A low level EIB driver for BCU 1.");
module_param (eib_major, int, 0);
MODULE_PARM_DESC (eib_major, "Major number for this device.");
MODULE_LICENSE ("GPL v2");

struct _eib_t;
typedef struct _eib_t eib_t;

typedef struct
{
  int port;
  int irq;
  long base;
  dev_t devid;
  int used;
  eib_t *eib;
  struct device dev;
} eib_config;

#define to_eib_config(d) container_of(d, eib_config, dev)

/* private minor number use count */
static eib_config devs[MAX_DEVICE_COUNT] = {
  {.port = 0x3f8,.irq = 4,.base = 115200},
  {.port = 0x2f8,.irq = 3,.base = 115200},
  {.port = 0x3e8,.irq = 4,.base = 115200},
  {.port = 0x2e8,.irq = 3,.base = 115200},
};

static struct platform_device *eib_device;
#ifdef HAVE_SIMPLE_CLASS
static struct class_simple *eib_class;
#else
static struct class *eib_class;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
static ssize_t
show_port (struct device *dev, char *buf)
#else
static ssize_t
show_port (struct device *dev, struct device_attribute *attr, char *buf)
#endif
{
  eib_config *conf = to_eib_config (dev);
  if (conf == NULL)
    return 0;

  return snprintf (buf, PAGE_SIZE, "%x\n", conf->port);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
static ssize_t
store_port (struct device *dev, const char *buf, size_t count)
#else
static ssize_t
store_port (struct device *dev, struct device_attribute *attr,
	    const char *buf, size_t count)
#endif
{
  eib_config *conf = to_eib_config (dev);
  if (conf == NULL)
    return 0;

  conf->port = (int) simple_strtol (buf, NULL, 16);
  return count;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
static ssize_t
show_irq (struct device *dev, char *buf)
#else
static ssize_t
show_irq (struct device *dev, struct device_attribute *attr, char *buf)
#endif
{
  eib_config *conf = to_eib_config (dev);
  if (conf == NULL)
    return 0;

  return snprintf (buf, PAGE_SIZE, "%d\n", conf->irq);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
static ssize_t
store_irq (struct device *dev, const char *buf, size_t count)
#else
static ssize_t
store_irq (struct device *dev, struct device_attribute *attr, const char *buf,
	   size_t count)
#endif
{
  eib_config *conf = to_eib_config (dev);
  if (conf == NULL)
    return 0;

  conf->irq = (int) simple_strtol (buf, NULL, 10);
  return count;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
static ssize_t
show_base (struct device *dev, char *buf)
#else
static ssize_t
show_base (struct device *dev, struct device_attribute *attr, char *buf)
#endif
{
  eib_config *conf = to_eib_config (dev);
  if (conf == NULL)
    return 0;

  return snprintf (buf, PAGE_SIZE, "%ld\n", conf->base);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
static ssize_t
store_base (struct device *dev, const char *buf, size_t count)
#else
static ssize_t
store_base (struct device *dev, struct device_attribute *attr,
	    const char *buf, size_t count)
#endif
{
  eib_config *conf = to_eib_config (dev);
  if (conf == NULL)
    return 0;

  conf->base = (int) simple_strtol (buf, NULL, 10);
  if (conf->base <= 0)
    conf->base = 115200;

  return count;
}

DEVICE_ATTR (port, S_IRUGO | S_IWUSR, show_port, store_port);
DEVICE_ATTR (irq, S_IRUGO | S_IWUSR, show_irq, store_irq);
DEVICE_ATTR (base, S_IRUGO | S_IWUSR, show_base, store_base);

static struct device_driver eib_driver = {
  .name = "eib",
  .bus = &platform_bus_type,
};

static ssize_t eib_read (struct file *file, char __user * buf, size_t count,
			 loff_t * offset);
static ssize_t eib_write (struct file *file, const char __user * buf,
			  size_t count, loff_t * offset);
static unsigned int eib_poll (struct file *file, poll_table * wait);
static int eib_ioctl (struct inode *inode, struct file *file,
		      unsigned int cmd, unsigned long arg);
static int eib_flush (struct file *file);
static int eib_fsync (struct file *file, struct dentry *d, int datasync);
static int eib_open (struct inode *inode, struct file *file);
static int eib_release (struct inode *inode, struct file *file);

static struct file_operations eib_fops = {
  .owner = THIS_MODULE,
  .read = eib_read,
  .write = eib_write,
  .poll = eib_poll,
  .ioctl = eib_ioctl,
  .open = eib_open,
  .flush = eib_flush,
  .fsync = eib_fsync,
  .release = eib_release,
};

static void
eib_device_release (struct device *dev)
{
  return;
}

static int __init
eib_init_module (void)
{
  int i;
  int retval;

  retval = register_chrdev (eib_major, "eib", &eib_fops);
  if (retval < 0)
    {
      return retval;
    }

  if (eib_major == 0)
    eib_major = retval;

  retval = driver_register (&eib_driver);
  if (retval != 0)
    goto unreg1;

  eib_device = platform_device_register_simple ("eib", -1, NULL, 0);
  if (IS_ERR (eib_device))
    {
      retval = PTR_ERR (eib_device);
      goto unreg2;
    }

#ifdef HAVE_SIMPLE_CLASS
  eib_class = class_simple_create (THIS_MODULE, "eib");
#else
  eib_class = class_create (THIS_MODULE, "eib");
#endif

  for (i = 0; i < MAX_DEVICE_COUNT; i++)
    {
      devs[i].devid = MKDEV (eib_major, i);
      devs[i].dev.parent = &eib_device->dev;
      devs[i].dev.release = eib_device_release;
      snprintf (devs[i].dev.bus_id, sizeof (devs[i].dev.bus_id), "eib%d", i);
      device_register (&devs[i].dev);
      device_create_file (&devs[i].dev, &dev_attr_port);
      device_create_file (&devs[i].dev, &dev_attr_irq);
      device_create_file (&devs[i].dev, &dev_attr_base);
#ifdef HAVE_SIMPLE_CLASS
      class_simple_device_add (eib_class, devs[i].devid, 0,
			       devs[i].dev.bus_id);
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,15)
      class_device_create (eib_class, devs[i].devid, NULL,
			   devs[i].dev.bus_id);
#else
      class_device_create (eib_class, NULL, devs[i].devid, NULL,
			   devs[i].dev.bus_id);
#endif
#endif
    }

  PDEBUG ("driver loaded (%d)", eib_major);

  return 0;

unreg2:
  driver_unregister (&eib_driver);
unreg1:
  unregister_chrdev (eib_major, "eib");
  return retval;
}

static void __exit
eib_cleanup_module (void)
{
  int i;
  for (i = 0; i < MAX_DEVICE_COUNT; i++)
    {
#ifdef HAVE_SIMPLE_CLASS
      class_simple_device_remove (devs[i].devid);
#else
      class_device_destroy (eib_class, devs[i].devid);
#endif
      device_remove_file (&devs[i].dev, &dev_attr_port);
      device_remove_file (&devs[i].dev, &dev_attr_irq);
      device_remove_file (&devs[i].dev, &dev_attr_base);
      device_unregister (&devs[i].dev);
    }
#ifdef HAVE_SIMPLE_CLASS
  class_simple_destroy (eib_class);
#else
  class_destroy (eib_class);
#endif
  platform_device_unregister (eib_device);
  driver_unregister (&eib_driver);
  unregister_chrdev (eib_major, "eib");

  PDEBUG ("successfully unloaded!\n");
}

module_init (eib_init_module);
module_exit (eib_cleanup_module);

typedef struct
{
  int port;
  int irq;
  long base;

  unsigned char old_dll;
  unsigned char old_dlh;
  unsigned char old_lcr;
  unsigned char old_mcr;
} eib_lowlevel;

enum eib_reason
{
  EIB_INT_DATA,
  EIB_INT_MSR,
  EIB_TIME_RESET,
  EIB_TIME_SEND,
  EIB_TIME_WRITE,
  EIB_NONE
};

typedef struct _eib_port
{
  void (*setRTS) (eib_lowlevel * dev);
  void (*clrRTS) (eib_lowlevel * dev);
  int (*getCTS) (eib_lowlevel * dev);
  void (*send) (eib_lowlevel * dev, uint8_t val);
  int (*recv) (eib_lowlevel * dev);
  void (*open_port) (eib_lowlevel * dev);
  void (*close_port) (eib_lowlevel * dev);
  enum eib_reason (*getserialint) (eib_lowlevel * dev);
  int (*send_finished) (eib_lowlevel * dev);
  int (*overrun) (eib_lowlevel * dev);
} eib_port;

static void
eib_setRTS (eib_lowlevel * dev)
{
  outb (UART_MCR_DTR | UART_MCR_RTS | UART_MCR_OUT2, dev->port + UART_MCR);
}

static void
eib_clrRTS (eib_lowlevel * dev)
{
  outb (UART_MCR_DTR | UART_MCR_OUT2, dev->port + UART_MCR);
}

static int
eib_getCTS (eib_lowlevel * dev)
{
  return inb (dev->port + UART_MSR) & UART_MSR_CTS;
}

static void
eib_send (eib_lowlevel * dev, uint8_t val)
{
  outb (val, dev->port + UART_TX);
}

static int
eib_recv (eib_lowlevel * dev)
{
  return inb (dev->port + UART_RX);
}

static int
eib_overrun (eib_lowlevel * dev)
{
  return inb (dev->port + UART_LSR) & UART_LSR_OE;
}

static int
eib_send_finished (eib_lowlevel * dev)
{
  return (inb (dev->port + UART_LSR) & UART_LSR_TEMT);
}

static enum eib_reason
eib_getserialint (eib_lowlevel * dev)
{
  int id = inb (dev->port + UART_IIR);

  if (~id & UART_IIR_NO_INT)
    {
      if ((id & UART_IIR_ID) == UART_IIR_RDI)
	return EIB_INT_DATA;

      if ((id & UART_IIR_MSI) == UART_IIR_MSI)
	if (inb (dev->port + UART_MSR) & UART_MSR_DCTS)
	  return EIB_INT_MSR;

    }
  return EIB_NONE;
}


/*********************************************************************
 ** Name    : init_serial
 ** Purpose : No serial driver used/this module programs the UART
 **           itself (Speed: 9600Baud, 8 Databits, 1 Stoppbit)
 ********************************************************************/
static void
eib_init_serial (eib_lowlevel * dev)
{
  int divisor = dev->base / 9600;
  outb (0x00, dev->port + UART_IER);
  dev->old_lcr = inb (dev->port + UART_LCR);	/* save old line control reg */

  outb (UART_LCR_DLAB, dev->port + UART_LCR);	/* set divisor latch access bit in line control reg */
  dev->old_dll = inb (dev->port + UART_DLL);	/* save divisor latch low */
  dev->old_dlh = inb (dev->port + UART_DLM);	/* save divisor latch high 
						 * 115200 / Divisorlatch = speed
						 */
  outb (divisor & 0xff, dev->port + UART_DLL);	/* write 0x0c to divisor latch low */
  outb ((divisor >> 8) & 0xff, dev->port + UART_DLM);	/* write 0x00 to divisor latch high 
							 * Speed : 9600
							 */

  outb (UART_LCR_WLEN8, dev->port + UART_LCR);	/* 8 Databits in the line control reg */
  outb (0x00, dev->port + UART_FCR);	/* disable Fifo control reg */
  dev->old_mcr = inb (dev->port + UART_MCR);	/* save modem control reg */

  outb (UART_MCR_DTR | UART_MCR_OUT2, dev->port + UART_MCR);
  /* Set RTS, Enable GP02 neccessary for UART ints */
  outb (UART_IER_MSI | UART_IER_RDI, dev->port + UART_IER);
  /* Enable modem status int and receiver data int */

  eib_recv (dev);
}

/*********************************************************************
 ** Name    : restore_serial
 ** Purpose : Restores the settings of the UART control regs from the
 **           variables explained above
 ********************************************************************/
static void
eib_restore_serial (eib_lowlevel * dev)
{
  outb (0x00, dev->port + UART_IER);
  outb (UART_LCR_DLAB, dev->port + UART_LCR);	/* enable divisor latch access */
  outb (dev->old_dll, dev->port + UART_DLL);
  outb (dev->old_dlh, dev->port + UART_DLM);
  outb (dev->old_lcr, dev->port + UART_LCR);
  outb (dev->old_mcr, dev->port + UART_MCR);
}


static eib_port eib_serial = {
  .setRTS = eib_setRTS,
  .clrRTS = eib_clrRTS,
  .getCTS = eib_getCTS,
  .send = eib_send,
  .recv = eib_recv,
  .open_port = eib_init_serial,
  .close_port = eib_restore_serial,
  .getserialint = eib_getserialint,
  .overrun = eib_overrun,
  .send_finished = eib_send_finished,
};

/* -- */

#define EIB_LOG(MSG, args...) if (0) do { struct timespec tv=CURRENT_TIME; PDEBUG("at %010ld.%09ld " MSG,tv.tv_sec,tv.tv_nsec,##args);} while(0)

#define MAX_READ_BUFFER 20	/* * 33 byte buffer size */
#define MAX_WRITE_BUFFER 5	/* * 33 byte buffer size */


// enable debug messsages

// maximum time in usec to wait between rx interrupt raised and byte fully send
#define MAX_TEMT_WAIT 1000

// minimum time in usec to wait between byte fully send and clear rts signal
#define WRITE_TEMT_CLRRTS_DELAY 100

// minimum time in usec to wait between clear rts and set rts signal
#define WRITE_CLRRTS_SETRTS_DELAY 80

// timeout for data block transfer
#define READ_WRITE_TIMEOUT 130000

// maximum time between to interrupts if not in idle state
#define MAX_TIME_BETWEEN_INT 80000

// minimum time to wait after reset without any respond to signals
#define MIN_RESET_WAIT_TIME 130000

// maximum time between to wait for start sending
#define MAX_INIT_SEND_TIMEOUT 100000

//
#define MAX_INIT_SEND_TIME 3000


static spinlock_t open_lock = SPIN_LOCK_UNLOCKED;



struct _eib_t
{
  int minor;			/* minor number */
  spinlock_t lock;
  spinlock_t irqlock;
  int shutdown;
  eib_port *iface;
  eib_lowlevel port;
  eib_lowlevel *dev;

  struct timer_list initate_send_timer_list;
  struct timer_list detect_reset_timer_list;
  struct timer_list write_timer_list;

  wait_queue_head_t read_queue;
  int read_head, read_free, read_next;
  unsigned char read_buffer[MAX_READ_BUFFER][33];
  spinlock_t read_lock;
  int read_pos;

  wait_queue_head_t write_queue;
  int write_head, write_free, write_next;
  unsigned char write_buffer[MAX_WRITE_BUFFER][33];
  signed char write_result[MAX_WRITE_BUFFER];
  spinlock_t write_lock;
  int write_pos;

  struct timespec last_int_tv;
  struct timespec start_tv;
  struct timespec send_tv;
  struct timespec reset_tv;	/* Time of the last reset */
  volatile int state;		/* actual state of the state-machine */
  volatile int initate_send;
  unsigned long warning_flag;
  int reset_count;

};

/*
 * Function headers
 *
 */
static int bitcount (unsigned char x);
static long time_difference (struct timespec *tv1, struct timespec *tv2);
static void add_initate_send_timer (struct file *file);
static void add_detect_reset_timer (struct file *file);
static void initate_send_timer (unsigned long data);
static void detect_reset_timer (unsigned long data);
static void write_timer (unsigned long data);
static irqreturn_t interrupt_handler (int irq, void *dev_id,
				      struct pt_regs *regs);
static void check_interrupt_pending (eib_t * dev);


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,11)
/* part of linux 2.6.11+ */
static unsigned long
usecs_to_jiffies (const unsigned int u)
{
  if (u > jiffies_to_usecs (MAX_JIFFY_OFFSET))
    return MAX_JIFFY_OFFSET;
#if HZ <= 1000000 && !(1000000 % HZ)
  return (u + (1000000 / HZ) - 1) / (1000000 / HZ);
#elif HZ > 1000000 && !(HZ % 1000000)
  return u * (HZ / 1000000);
#else
  return (u * HZ + 999999) / 1000000;
#endif
}
#endif


static void
add_initate_send_timer (struct file *file)
{
  eib_t *dev = (eib_t *) file->private_data;
  long t1, t2;
  if (dev->shutdown)
    {
      return;
    }
  t1 = MAX_INIT_SEND_TIME - time_difference (&dev->last_int_tv, NULL);
  t2 = MAX_INIT_SEND_TIMEOUT - time_difference (&dev->send_tv, NULL);
  if (t1 < 0)
    t1 = 0;
  if (t2 < 0)
    t2 = 0;

  if (t2 < t1)
    t1 = t2;
  mod_timer (&dev->initate_send_timer_list,
	     t1 ? jiffies + usecs_to_jiffies (t1) : jiffies + 1);
}

static void
add_detect_reset_timer (struct file *file)
{
  eib_t *dev = (eib_t *) file->private_data;
  long t1;
  if (dev->shutdown)
    {
      return;
    }
  if (dev->state == 7)
    {
      t1 = MIN_RESET_WAIT_TIME - time_difference (&dev->reset_tv, NULL);
      if (t1 < 0)
	t1 = 0;
    }
  else
    {
      long t2;
      t1 = READ_WRITE_TIMEOUT - time_difference (&dev->start_tv, NULL);
      t2 = MAX_TIME_BETWEEN_INT - time_difference (&dev->last_int_tv, NULL);
      if (t1 < 0)
	t1 = 0;
      if (t2 < 0)
	t2 = 0;
      if (t2 < t1)
	t1 = t2;
    }

  mod_timer (&dev->detect_reset_timer_list,
	     t1 ? jiffies + usecs_to_jiffies (t1) : jiffies + 1);
}

static void
add_write_timer (struct file *file, long timeout)
{
  eib_t *dev = (eib_t *) file->private_data;
  if (dev->shutdown)
    {
      return;
    }
  mod_timer (&dev->write_timer_list,
	     timeout ? jiffies + usecs_to_jiffies (timeout) : jiffies + 1);

}

/*********************************************************************
 ** Name    : bitcount
 ** Purpose : counts the 1s of a Byte 
 **
 **
 **
 ********************************************************************/
static int
bitcount (unsigned char x)
{
  int b;

  for (b = 0; x != 0; x >>= 1)
    if (x & 01)
      b++;

  return b;
}

/*********************************************************************
 ** Name    : time_difference
 ** Purpose : if second parameter == NULL => actual time will be used
 **
 **
 **
 ********************************************************************/
static long
time_difference (struct timespec *tv1, struct timespec *tv2)
{
  struct timespec foo;

  if (tv2 == NULL)
    {
      foo = CURRENT_TIME;
      tv2 = &foo;
    }

  return (tv2->tv_nsec - tv1->tv_nsec) / 1000 +	/* usecs   */
    1000000 * (tv2->tv_sec - tv1->tv_sec);	/* seconds * 1000000 */
}



/*********************************************************************
 ** Name    : eib_open
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static int
eib_open (struct inode *inode, struct file *file)
{
  eib_t *dev;
  unsigned long flags;
  int port;
  int irq;

  /* determine minor number (corresponds to serial port number) */
  unsigned int minor = iminor (inode);

  PDEBUG ("open %d", minor);
  /* don't handle higher minors as number of devices supported */
  if (minor >= MAX_DEVICE_COUNT)
    return -ENODEV;

  port = devs[minor].port;
  irq = devs[minor].irq;

  /* we only let one process open the device */
  /* kernel code is never preempted if holding a spinlock */
  spin_lock (&open_lock);
  if (devs[minor].used)
    {
      spin_unlock (&open_lock);
      return -EBUSY;
    }
  devs[minor].used = 1;
  spin_unlock (&open_lock);

  /* assign device-struct and initialize it */
  dev = (void *) kmalloc (sizeof (eib_t), GFP_KERNEL);
  if (!dev)
    goto outmem;
  memset (dev, 0, sizeof (eib_t));
  dev->dev = &dev->port;

  spin_lock_init (&dev->lock);
  spin_lock_init (&dev->irqlock);
  dev->shutdown = 0;

  file->private_data = dev;
  dev->minor = minor;
  dev->port.port = port;
  dev->port.irq = irq;
  dev->port.base = devs[minor].base;
  dev->iface = &eib_serial;

  if (!request_region (dev->port.port, 0x08, "eib"))
    {
      kfree (dev);
      goto outdev;
    }

  if (request_irq
      (dev->port.irq, interrupt_handler, SA_INTERRUPT | SA_SHIRQ, "eib",
       (void *) file) < 0)
    {
      release_region (dev->port.port, 0x08);
      kfree (dev);
      goto outdev;
    }

  PDEBUG ("resouces allocated %d", dev->minor);
  spin_lock_irqsave (&dev->lock, flags);
  init_timer (&dev->initate_send_timer_list);
  init_timer (&dev->detect_reset_timer_list);
  init_timer (&dev->write_timer_list);
  dev->initate_send_timer_list.function = initate_send_timer;
  dev->initate_send_timer_list.data = (unsigned long int) (void *) file;
  dev->detect_reset_timer_list.function = detect_reset_timer;
  dev->detect_reset_timer_list.data = (unsigned long int) (void *) file;
  dev->write_timer_list.function = write_timer;
  dev->write_timer_list.data = (unsigned long int) (void *) file;

  dev->state = 0;
  dev->initate_send = 0;
  dev->warning_flag = 0;
  dev->reset_count = 0;
  dev->last_int_tv = CURRENT_TIME;

  dev->read_head = 0;
  dev->read_free = 0;
  dev->read_next = 0;
  dev->read_pos = 0;
  spin_lock_init (&dev->read_lock);
  init_waitqueue_head (&dev->read_queue);

  dev->write_head = 0;
  dev->write_free = 0;
  dev->write_next = 0;
  dev->write_pos = 0;
  spin_lock_init (&dev->write_lock);
  init_waitqueue_head (&dev->write_queue);

  dev->iface->open_port (dev->dev);
  devs[minor].eib = dev;
  spin_unlock_irqrestore (&dev->lock, flags);

  return 0;

outmem:
  devs[minor].used = 0;
  return -ENOMEM;

outdev:
  devs[minor].used = 0;
  return -ENODEV;
}

/*********************************************************************
 ** Name    : eib_release
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static int
eib_release (struct inode *inode, struct file *file)
{
  unsigned long flags;
  eib_t *dev = file->private_data;

  PDEBUG ("close %d", dev->minor);
  dev->shutdown = 1;

  flush_scheduled_work ();
  synchronize_kernel ();

  spin_lock_irqsave (&dev->lock, flags);
  dev->iface->close_port (dev->dev);

  free_irq (dev->port.irq, (void *) file);
  release_region (dev->port.port, 0x08);

  del_timer_sync (&dev->initate_send_timer_list);
  del_timer_sync (&dev->write_timer_list);
  del_timer_sync (&dev->detect_reset_timer_list);

  devs[dev->minor].eib = 0;
  devs[dev->minor].used = 0;
  spin_unlock_irqrestore (&dev->lock, flags);

  PDEBUG ("closed %d", dev->minor);
  /* must (obviously enough) be last */
  kfree (dev);

  return 0;
}


/* -------   */

// EXPORT_NO_SYMBOLS;

#define EIB_BCU_VERSION 1	/* eib bcu version */

/*********************************************************************
 ** Name    : do_reset
 ** Purpose : Resets the serial port
 **
 **
 **
 ********************************************************************/
static void
do_reset1 (eib_t * dev, unsigned long reason, const char *msg, int line)
{

#ifdef EIB_DEBUG
  printk (KERN_NOTICE
	  "debug: do_reset: state=%d, reason=%lx, initate_send=%d, getCTS()=%d, %s at %d\n",
	  dev->state, reason, dev->initate_send,
	  dev->iface->getCTS (dev->dev), msg, line);
#endif

  dev->warning_flag |= reason;
  dev->state = 7;
  dev->iface->getserialint (dev->dev);
  dev->iface->recv (dev->dev);
  dev->iface->getCTS (dev->dev);	/* clear MSR Delta */
  dev->reset_tv = CURRENT_TIME;
  dev->iface->clrRTS (dev->dev);
}

#define do_reset(dev, MSG) do_reset1(dev, MSG , __LINE__)

#define ERR_1         0x00000001   , "no interrupt pending"
#define ERR_2         0x00000002   , "unexpected receive data interrupt"
#define ERR_4         0x00000004   , "unexpected receive data interrupt"
#define ERR_8         0x00000008   , "unexpected cts change to high"
#define ERR_10        0x00000010   , "read buffer overflow"
#define ERR_20        0x00000020   , "unexpected cts change to high"
#define ERR_40        0x00000040   , "unknown state"
#define ERR_80        0x00000080   , "overrun"
#define ERR_100       0x00000100   , "read forced"
#define ERR_200       0x00000200   , "unexpected receive data interrupt"
#define ERR_400       0x00000400   , "parity error"
#define ERR_800       0x00000800   , "unexpected receive data interrupt"
#define ERR_1000      0x00001000   , "exceed MAX_TEMT_WAIT"
#define ERR_2000      0x00002000   , "unexpected byte received"
#define ERR_4000      0x00004000   , "read forced && read buffer full"
#define ERR_8000      0x00008000   , "unexpected cts change"
#define ERR_10000     0x00010000   , "initate send with empty write buffer"
#define ERR_20000     0x00020000   , "bcu reset (info)"
#define ERR_40000     0x00040000   , "unexpected receive data interrupt"
#define ERR_80000     0x00080000   , "unexpected cts change to high"
#define ERR_100000    0x00100000   , "READ_WRITE_TIMEOUT"
#define ERR_200000    0x00200000   , "exceed MAX_TIME_BETWEEN_INT"
#define ERR_400000    0x00400000   , "clear write buffer"
#define ERR_800000    0x00800000   , "BCU reset"
#define ERR_10000000  0x10000000   , "INITIATESEND"


/*********************************************************************
 ** Name    : eib_read
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static ssize_t
eib_read (struct file *file, char __user * buf, size_t count, loff_t * offset)
{
  eib_t *dev = (eib_t *) file->private_data;
  int rb_tail, len;
  unsigned long flags;

retry:
  while (dev->read_next == dev->read_free)
    {
      if (file->f_flags & O_NONBLOCK)	/* read buffer empty */
	return -EAGAIN;
      wait_event_interruptible (dev->read_queue,
				!(dev->read_next == dev->read_free));

      if (signal_pending (current))
	return -ERESTARTSYS;
    }

  spin_lock_irqsave (&dev->read_lock, flags);


  if (dev->read_next == dev->read_free)
    {
      spin_unlock_irqrestore (&dev->read_lock, flags);
      goto retry;
    }

  rb_tail = dev->read_next;
  len = (dev->read_buffer[rb_tail][0] & 0x1f);

  if (count >= len)
    dev->read_next = (dev->read_next + 1) % MAX_READ_BUFFER;

  spin_unlock_irqrestore (&dev->read_lock, flags);

  if (count < len)
    return -ENOMSG;

  if (copy_to_user (buf, &dev->read_buffer[rb_tail][1], len))
    return -EFAULT;

  spin_lock_irqsave (&dev->read_lock, flags);

  dev->read_head = (dev->read_head + 1) % MAX_READ_BUFFER;

  spin_unlock_irqrestore (&dev->read_lock, flags);

  return len;
}


/*********************************************************************
 ** Name    : eib_write
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static ssize_t
eib_write (struct file *file,
	   const char __user * buf, size_t count, loff_t * offset)
{
  eib_t *dev = (eib_t *) file->private_data;
  int temp, wb_head;
  unsigned long flags;

  if (count > 32)
    return -EINVAL;
retry:
  while ((dev->write_free + 1) % MAX_WRITE_BUFFER == dev->write_head)
    {
      /* write buffer full */

      if (file->f_flags & O_NONBLOCK)
	return -EAGAIN;
      wait_event_interruptible (dev->write_queue,
				!((dev->write_free + 1) % MAX_WRITE_BUFFER ==
				  dev->write_head));

      if (signal_pending (current))
	return -ERESTARTSYS;
    }

  spin_lock_irqsave (&dev->write_lock, flags);

  if ((dev->write_free + 1) % MAX_WRITE_BUFFER == dev->write_head)
    {

      spin_unlock_irqrestore (&dev->write_lock, flags);
      goto retry;
    }


  wb_head = dev->write_free;
  dev->write_free = (dev->write_free + 1) % MAX_WRITE_BUFFER;

  spin_unlock_irqrestore (&dev->write_lock, flags);

  temp = (count & 0x1f) | 0x20;
  dev->write_buffer[wb_head][0] = temp | (bitcount (temp) % 2) * 0x80;
  if (copy_from_user (&dev->write_buffer[wb_head][1], buf, count))
    return -EFAULT;
  dev->write_result[wb_head] = 0;

  spin_lock_irqsave (&dev->write_lock, flags);

  dev->write_next = (dev->write_next + 1) % MAX_WRITE_BUFFER;
  dev->send_tv = CURRENT_TIME;
  add_initate_send_timer (file);

  spin_unlock_irqrestore (&dev->write_lock, flags);
  if (file->f_flags & O_NONBLOCK)
    return count & 0x1f;

  while (!dev->write_result[wb_head])
    {
      /* write buffer full */

      wait_event_interruptible (dev->write_queue, dev->write_result[wb_head]);

      if (signal_pending (current))
	return -EINTR;
    }

  if (dev->write_result[wb_head] > 0)
    return count & 0x1f;
  else
    return 0;
}




/*********************************************************************
 ** Name    : eib_poll
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static unsigned int
eib_poll (struct file *file, poll_table * wait)
{
  eib_t *dev = (eib_t *) file->private_data;
  unsigned int mask = 0;
  unsigned long flags;

  poll_wait (file, &dev->read_queue, wait);
  poll_wait (file, &dev->write_queue, wait);

  spin_lock_irqsave (&dev->read_lock, flags);
  if (dev->read_free != dev->read_next)
    mask |= POLLIN | POLLRDNORM;
  spin_unlock_irqrestore (&dev->read_lock, flags);

  spin_lock_irqsave (&dev->write_lock, flags);
  if ((dev->write_free + 1) % MAX_WRITE_BUFFER != dev->write_head)
    mask |= POLLOUT | POLLWRNORM;
  if (dev->write_head == dev->write_next)
    mask |= POLLPRI;
  spin_unlock_irqrestore (&dev->write_lock, flags);

  return mask;
}




/*********************************************************************
 ** Name    : eib_ioctl
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static int
eib_ioctl (struct inode *inode,
	   struct file *file, unsigned int cmd, unsigned long arg)
{
  unsigned long flags;
  eib_t *dev = (eib_t *) file->private_data;

  switch (cmd)
    {

    case EIB_GET_BCU_VERSION:
      return EIB_BCU_VERSION;

    case EIB_GET_DEVICE_INFORMATION:
      break;

    case EIB_RESET_DEVICE_STATISTIC:
      spin_lock_irqsave (&dev->lock, flags);

      dev->warning_flag = 0;

      spin_unlock_irqrestore (&dev->lock, flags);

      break;

    case EIB_CLEAR_WRITE_BUFFER:
      spin_lock_irqsave (&dev->write_lock, flags);

      del_timer (&dev->initate_send_timer_list);
      add_detect_reset_timer (file);
      dev->initate_send = 0;
      dev->warning_flag = 0;
      dev->write_head = dev->write_free;
      do_reset (dev, ERR_400000);

      spin_unlock_irqrestore (&dev->write_lock, flags);

      wake_up_interruptible_all (&dev->write_queue);
      break;

    case EIB_CLEAR_READ_BUFFER:
      spin_lock_irqsave (&dev->read_lock, flags);

      del_timer (&dev->initate_send_timer_list);
      add_detect_reset_timer (file);
      dev->warning_flag = 0;
      dev->read_free = dev->read_next;
      do_reset (dev, ERR_400000);

      spin_unlock_irqrestore (&dev->read_lock, flags);

      wake_up_interruptible_all (&dev->read_queue);
      break;

    case EIB_SINGLE_OPEN:
      break;

    default:
      return -ENOIOCTLCMD;
    }

  return 0;
}



/*********************************************************************
 ** Name    : eib_flush
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static int
eib_flush (struct file *file)
{
  eib_t *dev = (eib_t *) file->private_data;
  while (dev->write_head != dev->write_free)
    {				// write buffer not empty
      wait_event_interruptible (dev->write_queue,
				!(dev->write_head != dev->write_free));
      if (signal_pending (current))
	return -ERESTARTSYS;
    }

  return 0;
}

static int
eib_fsync (struct file *file, struct dentry *d, int datasync)
{
  eib_t *dev = (eib_t *) file->private_data;
  while (dev->write_head != dev->write_free)
    {				// write buffer not empty
      wait_event_interruptible (dev->write_queue,
				!(dev->write_head != dev->write_free));
      if (signal_pending (current))
	return -ERESTARTSYS;
    }

  return 0;
}

void
eib_handler (struct file *fp, enum eib_reason reason)
{
  eib_t *dev = (eib_t *) fp->private_data;
  unsigned long flags;
  int tmp;

  EIB_LOG ("Handler %d %d %d", reason, dev->state,
	   dev->iface->getCTS (dev->dev));

  switch (dev->state)
    {				/* State machine state */
    case 0:
      switch (reason)
	{
	case EIB_INT_DATA:
	  do_reset (dev, ERR_2);	/* unexpected receive data interrupt */
	  break;
	case EIB_INT_MSR:
	  if (dev->iface->getCTS (dev->dev))
	    {
	      if (dev->initate_send)
		{		/* write to bcu */
		  dev->iface->send (dev->dev,
				    dev->write_buffer[dev->write_head][0]);
		  dev->initate_send = 0;
		  dev->write_pos++;
		  dev->state = 5;

		  break;
		}
	      else
		{		// read
		  add_detect_reset_timer (fp);
		  if ((dev->read_free + 1) % MAX_READ_BUFFER ==
		      dev->read_head)
		    {
		      do_reset (dev, ERR_10);	/* read buffer overflow */
		    }
		  else
		    {
		      dev->start_tv = CURRENT_TIME;
		      dev->iface->setRTS (dev->dev);
		      dev->iface->send (dev->dev, 0xFF);
		      dev->state = 2;
		    }
		}
	    }			// else !getCTS() -> state=0
	  break;
	case EIB_TIME_RESET:
	  break;
	case EIB_TIME_SEND:
	  break;
	case EIB_TIME_WRITE:
	  do_reset (dev, ERR_40);	/* unknown state */
	  break;
	case EIB_NONE:
	  break;
	}
      break;
    case 1:
      switch (reason)
	{
	case EIB_INT_DATA:
	  do_reset (dev, ERR_200);	// unexpected receive data interrupt
	  break;
	case EIB_INT_MSR:
	  if (dev->iface->getCTS (dev->dev))
	    {
	      dev->iface->setRTS (dev->dev);
	      dev->iface->send (dev->dev, 0x00);
	      dev->state = 2;
	      add_detect_reset_timer (fp);
	    }			// else !getCTS() -> state=1
	  break;
	case EIB_TIME_RESET:
	  break;
	case EIB_TIME_SEND:
	  break;
	case EIB_TIME_WRITE:
	  do_reset (dev, ERR_40);	/* unknown state */
	  break;
	case EIB_NONE:
	  break;
	}
      break;
    case 2:
      switch (reason)
	{
	case EIB_INT_DATA:
	  tmp = dev->iface->recv (dev->dev);
	  dev->read_buffer[dev->read_free][dev->read_pos++] = tmp;
	  if (dev->iface->overrun (dev->dev))
	    {
	      do_reset (dev, ERR_80);	// overrun
	      break;
	    }
	  if (tmp == 0xA0 && dev->read_pos == 1)
	    {
	      do_reset (dev, ERR_800000);	// BCU reset
	      break;
	    }

	  if (tmp == 0xFF && dev->read_pos == 1)
	    {
	      PDEBUG ("No data");
	      dev->state = 0;
	      check_interrupt_pending (dev);
	      dev->iface->clrRTS (dev->dev);
	      dev->read_pos = 0;

	      dev->last_int_tv = CURRENT_TIME;
	      break;
	    }

	  if (dev->read_pos == 1 && !(((tmp & 0x60) == 0x20) &&
				      ((tmp & 0x80) ==
				       (bitcount (tmp & 0x7f) % 2) * 0x80)))
	    {
	      do_reset (dev, ERR_400);	// parity error
	      break;
	    }

	  dev->state = 3;
	  break;
	case EIB_INT_MSR:
	  if (dev->iface->getCTS (dev->dev))
	    do_reset (dev, ERR_20);	/* unexpected cts change to high */
	  break;
	case EIB_TIME_RESET:
	  break;
	case EIB_TIME_SEND:
	  break;
	case EIB_TIME_WRITE:
	  do_reset (dev, ERR_40);	/* unknown state */
	  break;
	case EIB_NONE:
	  break;
	}
      break;
    case 3:
      switch (reason)
	{
	case EIB_INT_DATA:
	  do_reset (dev, ERR_4);	/* unexpected receive data interrupt */
	  break;
	case EIB_INT_MSR:
	  if (dev->iface->getCTS (dev->dev))
	    {
	      do_reset (dev, ERR_8);	/* unexpected cts change to high */
	      break;
	    }
	  if (dev->read_pos ==
	      (dev->read_buffer[dev->read_free][0] & 0x1f) + 1)
	    {			// read finished

	      if (dev->read_buffer[dev->read_free][0] == 0xa0)
		dev->warning_flag |= 0x20000;	// read bcu reset
	      else
		{
		  dev->reset_count = 0;
		  dev->read_free = (dev->read_free + 1) % MAX_READ_BUFFER;
		  wake_up_interruptible_all (&dev->read_queue);
		}

	      dev->read_pos = 0;

	      spin_lock_irqsave (&dev->lock, flags);

	      if (dev->write_head != dev->write_next)
		{
		  dev->send_tv = CURRENT_TIME;
		  add_initate_send_timer (fp);
		}
	      dev->state = 0;
	    }
	  else
	    {
	      spin_lock_irqsave (&dev->lock, flags);
	      dev->state = 1;
	    }
	  check_interrupt_pending (dev);
	  dev->iface->clrRTS (dev->dev);

	  spin_unlock_irqrestore (&dev->lock, flags);

	  dev->last_int_tv = CURRENT_TIME;
	  break;
	case EIB_TIME_RESET:
	  break;
	case EIB_TIME_SEND:
	  break;
	case EIB_TIME_WRITE:
	  break;
	case EIB_NONE:
	  break;
	}
      break;
    case 4:
      switch (reason)
	{
	case EIB_INT_DATA:
	  do_reset (dev, ERR_800);	/* unexpected receive data interrupt */
	  break;
	case EIB_INT_MSR:
	  if (dev->iface->getCTS (dev->dev))
	    {
	      dev->iface->send (dev->dev,
				dev->write_buffer[dev->write_head][dev->
								   write_pos++]);
	      dev->state = 5;
	      add_detect_reset_timer (fp);
	    }			// else !getCTS() -> state=4
	  break;
	case EIB_TIME_RESET:
	  break;
	case EIB_TIME_SEND:
	  break;
	case EIB_TIME_WRITE:
	  do_reset (dev, ERR_40);	/* unknown state */
	  break;
	case EIB_NONE:
	  break;
	}
      break;
    case 5:
      switch (reason)
	{
	case EIB_INT_DATA:
	  tmp = dev->iface->recv (dev->dev);

	  dev->state = 6;

	  if (dev->iface->overrun (dev->dev))
	    {
	      do_reset (dev, ERR_80);	/* overun */
	      break;
	    }
	  if (dev->write_pos == 1 && tmp == 0xA0)
	    {
	      do_reset (dev, ERR_800000);	// BCU reset
	      break;
	    }

	  if (dev->write_pos == 1 && tmp != 0xff)
	    {

	      dev->warning_flag |= 0x100;	/* read forced */
	      dev->write_pos = 0;
	      if ((dev->read_free + 1) % MAX_READ_BUFFER == dev->read_head)
		{
		  do_reset (dev, ERR_4000);	/* read forced && read buffer full */

		  break;
		}

	      if (((tmp & 0x60) != 0x20) ||
		  ((tmp & 0x80) != (bitcount (tmp & 0x7f) % 2) * 0x80))
		{
		  do_reset (dev, ERR_400);	/* parity error */

		  break;
		}
	      dev->read_buffer[dev->read_free][dev->read_pos++] = tmp;

	      dev->state = 3;
	      break;
	    }
	  if (dev->write_pos > 1 && tmp != 0x00)
	    {
	      do_reset (dev, ERR_2000);	/* unexpect byte received */
	      break;
	    }
	  add_write_timer (fp, MAX_TEMT_WAIT + WRITE_TEMT_CLRRTS_DELAY);
	  break;
	case EIB_INT_MSR:
	  do_reset (dev, ERR_8000);	/* unexpected cts change */
	  break;
	case EIB_TIME_RESET:
	  break;
	case EIB_TIME_SEND:
	  break;
	case EIB_TIME_WRITE:
	  do_reset (dev, ERR_40);	/* unknown state */
	  break;
	case EIB_NONE:
	  break;
	}
      break;
    case 6:
      switch (reason)
	{
	case EIB_INT_DATA:
	  do_reset (dev, ERR_4);	/* unexpected receive data interrupt */
	  break;
	case EIB_INT_MSR:
	  if (dev->iface->getCTS (dev->dev))
	    do_reset (dev, ERR_8);	/* unexpected cts change to high */
	  break;
	case EIB_TIME_RESET:
	  break;
	case EIB_TIME_SEND:
	  break;
	case EIB_TIME_WRITE:
	  if (!dev->iface->send_finished (dev->dev))
	    {
	      do_reset (dev, ERR_1000);	// exceed MAX_TEMT_WAIT
	      return;
	    }

	  if (dev->write_pos ==
	      (dev->write_buffer[dev->write_head][0] & 0x1f) + 1)
	    {			// write finished

	      dev->write_result[dev->write_head] = 1;
	      dev->reset_count = 0;
	      dev->write_head = (dev->write_head + 1) % MAX_WRITE_BUFFER;
	      wake_up_interruptible_all (&dev->write_queue);
	      dev->write_pos = 0;

	      if (dev->write_head != dev->write_next)
		{
		  dev->send_tv = CURRENT_TIME;
		  add_initate_send_timer (fp);
		}

	      spin_lock_irqsave (&dev->lock, flags);


	      dev->state = 0;

	      check_interrupt_pending (dev);
	      dev->iface->clrRTS (dev->dev);
	      spin_unlock_irqrestore (&dev->lock, flags);
	    }
	  else
	    {
	      check_interrupt_pending (dev);

	      dev->iface->clrRTS (dev->dev);

	      spin_lock_irqsave (&dev->lock, flags);
	      dev->state = 8;
	      add_write_timer (fp, WRITE_CLRRTS_SETRTS_DELAY);
	      spin_unlock_irqrestore (&dev->lock, flags);
	    }

	  dev->last_int_tv = CURRENT_TIME;
	  break;
	case EIB_NONE:
	  break;
	}
      break;
    case 7:
      switch (reason)
	{
	case EIB_INT_DATA:
	  dev->iface->recv (dev->dev);
	  break;
	case EIB_INT_MSR:
	  dev->iface->getCTS (dev->dev);
	  break;
	case EIB_TIME_RESET:
	  break;
	case EIB_TIME_SEND:
	  break;
	case EIB_TIME_WRITE:
	  break;
	case EIB_NONE:
	  break;
	}
      break;
    case 8:
      switch (reason)
	{
	case EIB_INT_DATA:
	  do_reset (dev, ERR_4);	/* unexpected receive data interrupt */
	  break;
	case EIB_INT_MSR:
	  if (dev->iface->getCTS (dev->dev))
	    do_reset (dev, ERR_8);	/* unexpected cts change to high */
	  del_timer (&dev->write_timer_list);
	  goto rt2;
	  break;
	case EIB_TIME_RESET:
	  break;
	case EIB_TIME_SEND:
	  break;
	case EIB_TIME_WRITE:
	rt2:
	  spin_lock_irqsave (&dev->lock, flags);

	  dev->state = 4;

	  dev->iface->setRTS (dev->dev);
	  check_interrupt_pending (dev);
	  spin_unlock_irqrestore (&dev->lock, flags);

	  dev->last_int_tv = CURRENT_TIME;
	  break;
	case EIB_NONE:
	  break;
	}
      break;
    default:
      do_reset (dev, ERR_40);	// unknown state    
    }
}




/*********************************************************************
 ** Name    : initate_send_timer
 ** Purpose : ?
 **
 **
 **
 ********************************************************************/
static void
initate_send_timer (unsigned long dev_id)
{
  struct file *fp = (struct file *) dev_id;
  eib_t *dev = (eib_t *) fp->private_data;
  unsigned long flags;

  EIB_LOG ("Initiate_Send %d %d", dev->state, dev->iface->getCTS (dev->dev));

  if (dev->write_head == dev->write_next)
    {
      EIB_PRINT ("initate send with empty buffer");
      dev->warning_flag |= 0x10000;	// initate send with empty write buffer
      return;
    }

  spin_lock_irqsave (&dev->lock, flags);

  if (dev->state == 0 && !dev->iface->getCTS (dev->dev) &&
      time_difference (&dev->last_int_tv, NULL) > MAX_INIT_SEND_TIME)
    {

      dev->initate_send = 1;
      dev->start_tv = CURRENT_TIME;
      dev->last_int_tv = CURRENT_TIME;

      dev->iface->setRTS (dev->dev);
      add_detect_reset_timer (fp);

      spin_unlock_irqrestore (&dev->lock, flags);

      return;
    }
  if (time_difference (&dev->send_tv, NULL) > MAX_INIT_SEND_TIMEOUT)
    {
      EIB_PRINT ("timeout initiate send");

      do_reset (dev, ERR_10000000);
      add_detect_reset_timer (fp);

      spin_unlock_irqrestore (&dev->lock, flags);

      return;
    }

  add_initate_send_timer (fp);

  spin_unlock_irqrestore (&dev->lock, flags);
}


static void
write_timer (unsigned long dev_id)
{
  struct file *fp = (struct file *) dev_id;

  eib_handler (fp, EIB_TIME_WRITE);
}

/*********************************************************************
 ** Name    : detect_reset_timer
 ** Purpose : ?
 **
 **
 **
 ********************************************************************/
static void
detect_reset_timer (unsigned long dev_id)
{
  struct file *fp = (struct file *) dev_id;
  eib_t *dev = (eib_t *) fp->private_data;
  unsigned long flags;

  spin_lock_irqsave (&dev->lock, flags);

  EIB_LOG ("detect_reset %d %d", dev->state, dev->iface->getCTS (dev->dev));

  if (dev->state == 0 && !dev->initate_send && !dev->iface->getCTS (dev->dev))
    {				// idle

      spin_unlock_irqrestore (&dev->lock, flags);

      return;
    }
  else if (dev->state == 7)
    {				// reset
      if (time_difference (&dev->reset_tv, NULL) > MIN_RESET_WAIT_TIME)
	{
	  dev->last_int_tv = CURRENT_TIME;
	  if (dev->write_head != dev->write_next && dev->reset_count > 10)
	    {
	      dev->reset_count = 0;
	      PDEBUG ("discard packet");
	      dev->write_result[dev->write_head] = -1;
	      dev->write_head = (dev->write_head + 1) % MAX_WRITE_BUFFER;
	      wake_up_interruptible_all (&dev->write_queue);
	    }
	  if (dev->write_head != dev->write_next)
	    {
	      dev->send_tv = CURRENT_TIME;
	      add_initate_send_timer (fp);
	    }
	  dev->read_pos = 0;
	  dev->write_pos = 0;
	  dev->state = 0;
	  dev->reset_count++;
	  spin_unlock_irqrestore (&dev->lock, flags);

	  return;
	}
    }
  else
    {				// read || write
      if (time_difference (&dev->start_tv, NULL) > READ_WRITE_TIMEOUT)
	do_reset (dev, ERR_100000);	// read write timeout
      else if (time_difference (&dev->last_int_tv, NULL) >
	       MAX_TIME_BETWEEN_INT)
	do_reset (dev, ERR_200000);	// max time between int exceeded
    }
  add_detect_reset_timer (fp);

  spin_unlock_irqrestore (&dev->lock, flags);
}




/*********************************************************************
 ** Name    : check_interrupt_pending
 ** Purpose : 
 **
 **
 **
 ********************************************************************/
static void
check_interrupt_pending (eib_t * dev)
{
  switch (dev->iface->getserialint (dev->dev))
    {
    case EIB_INT_DATA:
      do_reset (dev, ERR_40000);	// unexpected receive data interrupt
      break;
    case EIB_INT_MSR:
      if (dev->iface->getCTS (dev->dev))
	do_reset (dev, ERR_80000);	// unexpected cts change to high
      break;
    default:
      break;
    }
}

/*********************************************************************
 ** Name    : interrupt_handler
 ** Purpose : Performs the state machine
 **
 **
 **
 ********************************************************************/
static irqreturn_t
interrupt_handler (int irq, void *dev_id, struct pt_regs *regs)
{
  struct file *fp = (struct file *) dev_id;
  eib_t *dev = (eib_t *) fp->private_data;

  spin_lock (&dev->irqlock);

  eib_handler (fp, dev->iface->getserialint (dev->dev));

  if (dev->state != 0 && dev->state != 7 &&
      time_difference (&dev->start_tv, NULL) > READ_WRITE_TIMEOUT)
    do_reset (dev, ERR_100000);	/* read write timeout */

  dev->last_int_tv = CURRENT_TIME;
  spin_unlock (&dev->irqlock);
  return IRQ_HANDLED;
}
